from .. import views

activation_complete = views.ActivationComplete.as_view()
registration_complete = views.RegistrationComplete.as_view()
registration_closed = views.RegistrationClosed.as_view()
register = views.Register.as_view()


class NoNotificationRegistration(views.Register):
    def send_notification(self):
        return
register_with_no_notification = NoNotificationRegistration.as_view()

register_but_closed = views.Register.as_view(registration_closed=True)
activate = views.Activate.as_view()


register_expired = views.Register.as_view(
    notification_template_name=('le_social/registration/'
                                'expired_activation_email.txt'),
)


class ExpiringActivate(views.Activate):
    expires_in = -1  # In the past
activate_expired = ExpiringActivate.as_view()



from django.contrib.auth.models import User
from django.core import mail
from django.core.urlresolvers import reverse
from django.test import TestCase


class RegistrationTest(TestCase):
    urls = 'le_social.registration.tests.urls'
    valid_data = {
        'username': 'brutasse',
        'email': 'foo@example.com',
        'password1': 'foo',
        'password2': 'foo',
    }

    def test_static_pages(self):
        """Static registration-related pages"""
        for name in ('activation_complete', 'complete', 'closed'):
            url = reverse('registration_%s' % name)
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)

    def test_register(self):
        self.assertEqual(User.objects.count(), 0)
        url = reverse('registration_register')
        response = self.client.get(url)
        self.assertContains(response, '<form ')
        data = {
            'username': '")*(@@!!',
            'email': 'yo dawg',
            'password1': 'foo',
            'password2': 'bar',
        }
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 200)
        self.assertFormError(response, 'form', 'username',
                             'The username must contain only letters, numbers '
                             'and underscores.')
        try:
            self.assertFormError(response, 'form', 'email',
                                 'Enter a valid email address.')
        except AssertionError:
            self.assertFormError(response, 'form', 'email',
                                 'Enter a valid e-mail address.')
        self.assertFormError(response, 'form', None,
                             "The two passwords didn't match.")
        response = self.client.post(url, self.valid_data, follow=True)
        self.assertContains(response, "Registration complete")
        self.assertEqual(len(response.redirect_chain), 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(User.objects.count(), 1)

    def test_strip_subject_line(self):
        url = reverse('registration_register')
        response = self.client.post(url, {
            'username': 'test',
            'email': 'test@example.com',
            'password1': 'foo',
            'password2': 'foo',
        })
        self.assertEqual(response.status_code, 302)
        self.assertEqual(len(mail.outbox), 1)
        # This raises BadHeaderError if header is not properly stripped
        mail.outbox[0].message()

    def test_register_with_no_notification(self):
        url = reverse('registration_register_with_notification')
        self.assertEqual(len(mail.outbox), 0)
        response = self.client.post(url, self.valid_data, follow=True)
        self.assertContains(response, "Registration complete")
        self.assertEqual(len(response.redirect_chain), 1)
        self.assertEqual(len(mail.outbox), 0)

    def test_activate(self):
        url = reverse('registration_register')
        response = self.client.post(url, self.valid_data)

        self.assertFalse(User.objects.get().is_active)
        url = mail.outbox[0].body.split('testserver')[1].split('\n')[0]
        response = self.client.get(url, follow=True)
        self.assertContains(response, "Account successfully activated")
        self.assertEqual(len(response.redirect_chain), 1)

    def test_closed_registration(self):
        url = reverse('registration_register_but_closed')
        response = self.client.get(url, follow=True)
        self.assertContains(response, "Registration is closed")
        self.assertEqual(len(response.redirect_chain), 1)

    def test_activation_key_expired(self):
        """
        Registration has expired
        """
        url = reverse('registration_register_expired')
        response = self.client.post(url, self.valid_data)
        url = mail.outbox[0].body.split('testserver')[1].split('\n')[0]
        response = self.client.get(url)
        self.assertContains(response, 'Invalid')



from __future__ import absolute_import

import twitter

from django.http import HttpResponse

from .. import views

kwargs = {'consumer_key': 'key',
          'consumer_secret': 'secret'}

authorize = views.Authorize.as_view(**kwargs)


class Callback(views.Callback):
    def error(self, message, exception=None):
        return HttpResponse(message)

    def success(self, auth):
        api = twitter.Twitter(auth=auth)
        user = api.account.verify_credentials()
        return HttpResponse(user['screen_name'])
callback = Callback.as_view(**kwargs)



from django.conf import settings
from django.contrib.auth.models import User
from django.contrib.sites.models import RequestSite
from django.core import signing
from django.core.mail import send_mail
from django.core.urlresolvers import reverse_lazy
from django.shortcuts import redirect
from django.template.loader import render_to_string
from django.utils.encoding import force_text
from django.views import generic

from .forms import RegistrationForm


class ActivationComplete(generic.TemplateView):
    template_name = 'le_social/registration/activation_complete.html'


class Activate(generic.TemplateView):
    template_name = 'le_social/registration/activate.html'
    success_url = reverse_lazy('registration_activation_complete')
    expires_in = 60 * 60 * 24 * 30  # 30 days

    def dispatch(self, request, *args, **kwargs):
        try:
            self.activation_key = signing.loads(kwargs['activation_key'],
                                                max_age=self.get_expires_in(),
                                                salt='le_social.registration')
        except signing.BadSignature:
            return super(Activate, self).dispatch(request, *args, **kwargs)
        self.activate()
        return redirect(self.get_success_url())

    def get_expires_in(self):
        return self.expires_in

    def get_success_url(self):
        return force_text(self.success_url)

    def activate(self):
        User.objects.filter(pk=self.activation_key).update(is_active=True)


class Register(generic.FormView):
    closed_url = reverse_lazy('registration_closed')
    form_class = RegistrationForm
    registration_closed = False
    success_url = reverse_lazy('registration_complete')
    template_name = 'le_social/registration/register.html'
    notification_template_name = 'le_social/registration/activation_email.txt'
    notification_subject_template_name = ('le_social/registration/'
                                          'activation_email_subject.txt')

    def dispatch(self, request, *args, **kwargs):
        if self.get_registration_closed():
            return redirect(self.get_closed_url())
        return super(Register, self).dispatch(request, *args, **kwargs)

    def get_registration_closed(self):
        return self.registration_closed

    def get_closed_url(self):
        return force_text(self.closed_url)

    def form_valid(self, form):
        self.user = form.save()
        self.activation_key = signing.dumps(self.user.pk,
                                            salt='le_social.registration')
        self.send_notification()
        return super(Register, self).form_valid(form)

    def get_notification_context(self):
        return {
            'user': self.user,
            'activation_key': self.activation_key,
            'site': RequestSite(self.request),
        }

    def send_notification(self):
        context = self.get_notification_context()
        send_mail(
            render_to_string(self.notification_subject_template_name,
                             context).strip(),
            render_to_string(self.notification_template_name, context),
            settings.DEFAULT_FROM_EMAIL,
            [self.user.email],
        )


class RegistrationComplete(generic.TemplateView):
    template_name = 'le_social/registration/registration_complete.html'


class RegistrationClosed(generic.TemplateView):
    template_name = 'le_social/registration/registration_closed.html'



import base64
import operator
import time

from hashlib import md5

try:
    from openid.association import Association as OIDAssociation
    from openid.store import nonce
    from openid.store.interface import OpenIDStore
except ImportError:
    from django.core.exceptions import ImproperlyConfigured
    raise ImproperlyConfigured(
        "python-openid is required to use le_social.openid"
    )

from django.conf import settings
from django.db.models import Q

from .models import Association, Nonce


class DjangoOpenIDStore(OpenIDStore):
    def __init__(self):
        self.max_nonce_age = 6 * 60 * 60  # Six hours

    def storeAssociation(self, server_url, association):
        Association.objects.create(
            server_url=server_url,
            handle=association.handle,
            secret=base64.encodestring(association.secret),
            issued=association.issued,
            lifetime=association.lifetime,
            assoc_type=association.assoc_type,
        )

    def getAssociation(self, server_url, handle=None):
        assocs = []
        if handle is not None:
            assocs = Association.objects.filter(
                server_url=server_url, handle=handle,
            )
        else:
            assocs = Association.objects.filter(server_url=server_url)
        if not assocs:
            return None
        associations = []
        expired = []
        for assoc in assocs:
            association = OIDAssociation(
                assoc.handle, base64.decodestring(assoc.secret), assoc.issued,
                assoc.lifetime, assoc.assoc_type,
            )
            if association.getExpiresIn() == 0:
                expired.append(assoc)
            else:
                associations.append((association.issued, association))

        for assoc in expired:
            assoc.delete()
        if not associations:
            return None
        associations.sort()
        return associations[-1][1]

    def removeAssociation(self, server_url, handle):
        assocs = list(Association.objects.filter(
            server_url=server_url, handle=handle,
        ))
        assocs_exist = len(assocs) > 0
        for assoc in assocs:
            assoc.delete()
        return assocs_exist

    def useNonce(self, server_url, timestamp, salt):
        if abs(timestamp - time.time()) > nonce.SKEW:
            return False

        query = [
            Q(server_url__exact=server_url),
            Q(timestamp__exact=timestamp),
            Q(salt__exact=salt),
        ]
        try:
            Nonce.objects.get(reduce(operator.and_, query))
        except Nonce.DoesNotExist:
            Nonce.objects.create(
                server_url=server_url,
                timestamp=timestamp,
                salt=salt,
            )
            return True
        return False

    def cleanupNonces(self, _now=None):
        if _now is None:
            _now = int(time.time())
        limit = _now - nonce.SKEW
        expired = Nonce.objects.filter(timestamp__lt=limit)
        count = expired.count()
        if count:
            expired.delete()
        return count

    def cleanupAssociations(self):
        now = int(time.time())
        expired = Association.objects.extra(
            where=['issued + lifetime < %d' % now])
        count = expired.count()
        if count:
            expired.delete()
        return count

    def getAuthKey(self):
        return md5(settings.SECRET_KEY).hexdigest()[:self.AUTH_KEY_LEN]

    def isDumb(self):
        return False



from django.conf.urls import patterns, url

from . import views


urlpatterns = patterns(
    '',
    url(r'^openid/$', views.begin, name='openid_begin'),
    url(r'^openid/complete/$', views.callback, name='openid_callback'),
)



from django.conf.urls import patterns, url

from . import views

urlpatterns = patterns(
    '',
    url(r'^activate/complete/$', views.activation_complete,
        name='registration_activation_complete'),

    url(r'^activate/(?P<activation_key>.+)/$', views.activate,
        name='registration_activate'),

    url(r'^activate-expired/(?P<activation_key>.+)/$', views.activate_expired,
        name='registration_activate_expired'),

    url(r'^register/$', views.register,
        name='registration_register'),

    url(r'^register-no-notify/$', views.register_with_no_notification,
        name='registration_register_with_notification'),

    url(r'^register-closed/$', views.register_but_closed,
        name='registration_register_but_closed'),

    url(r'^register-expired/$', views.register_expired,
        name='registration_register_expired'),

    url(r'^register/complete/$', views.registration_complete,
        name='registration_complete'),

    url(r'^register/closed/$', views.registration_closed,
        name='registration_closed'),
)



from django.conf.urls import patterns, url

from . import views


urlpatterns = patterns(
    '',
    url(r'^oauth/authorize/$', views.authorize, name='authorize'),
    url(r'^oauth/callback/$', views.callback, name='callback'),
)



from django.db import models
from django.utils.translation import ugettext_lazy as _


class Nonce(models.Model):
    server_url = models.CharField(_('Server URL'), max_length=255)
    timestamp = models.IntegerField(_('Timestamp'))
    salt = models.CharField(_('Salt'), max_length=40)

    def __unicode__(self):
        return u'Nonce: %s' % self.pk


class Association(models.Model):
    server_url = models.CharField(_('Server URL'), max_length=2047)
    handle = models.CharField(_('OpenID handle'), max_length=255)
    secret = models.CharField(_('OpenID secret'), max_length=255)
    issued = models.IntegerField(_('Issued'))
    lifetime = models.IntegerField(_('Lifetime'))
    assoc_type = models.CharField(_('Association type'), max_length=64)

    def __unicode__(self):
        return u'Association: %s, %s' % (self.server_url, self.handle)



from django import forms
from django.contrib.auth.models import User
from django.utils.translation import ugettext_lazy as _


class RegistrationForm(forms.Form):
    username = forms.RegexField(
        regex=r'^[\w.@+-]+$', max_length=30, label=_('Username'),
        error_messages={'invalid': _('The username must contain only letters, '
                                     'numbers and underscores.')},
    )
    email = forms.EmailField(label=_('Email'))
    password1 = forms.CharField(label=_('Password'),
                                widget=forms.PasswordInput)
    password2 = forms.CharField(label=_('Password (again)'),
                                widget=forms.PasswordInput)

    def clean(self):
        data = self.cleaned_data
        if ('password1' in data and 'password2' in data):
            if data['password1'] != data['password2']:
                raise forms.ValidationError(
                    _("The two passwords didn't match."),
                )
        return self.cleaned_data

    def save(self):
        user = User.objects.create_user(
            self.cleaned_data['username'],
            self.cleaned_data['email'],
            self.cleaned_data['password1'],
        )
        user.is_active = False
        user.save()
        return user



from openid.consumer import consumer
from openid.message import Message

from django.core.urlresolvers import reverse
from django.test import TestCase

from .. import utils


def discover_extensions(openid_url):
    """
    Don't bother with extensions for tests.
    """
    return False, False


class OpenidRequest(object):
    """
    Fake OpenID request
    """
    def __init__(self, url):
        self.url = url

    def addExtension(self, sreg_request):
        pass

    def redirectURL(self, trust_root, return_url):
        return trust_root[:-1] + return_url


class OpenidResponse(object):
    """
    Fake OpenID response
    """
    def __init__(self, status, message):
        self.status = status
        self.message = message
        self.identity_url = 'http://bruno.renie.fr'
        self.signed_fields = None

    def getSignedNS(self, uri):
        return self.message.getArgs(uri)

    def extensionResponse(self, namespace_uri, require_signed):
        if require_signed:
            return self.getSignedNS(namespace_uri)
        else:
            return self.message.getArgs(namespace_uri)


class Consumer(object):
    """
    Fake OpenID consumer
    """
    def __init__(self, session, store_class):
        self.session = session
        self.store_class = store_class

    def begin(self, url):
        return OpenidRequest(url)

    def complete(self, query, return_url):
        if not query:
            return OpenidResponse(consumer.FAILURE, 'Invalid openid.mode')
        return OpenidResponse(consumer.SUCCESS, Message())


class OpenidTest(TestCase):
    urls = 'le_social.openid.tests.urls'

    def setUp(self):
        self._real_consumer = consumer.Consumer
        consumer.Consumer = Consumer

        self._real_discover_extensions = utils.discover_extensions
        utils.discover_extensions = discover_extensions

    def tearDown(self):
        consumer.Consumer = self._real_consumer
        utils.discover_extensions = self._real_discover_extensions

    def test_openid_assoc(self):
        url = reverse('openid_begin')
        response = self.client.get(url)
        self.assertContains(response, '<form ')
        data = {'openid_url': 'http://bruno.renie.fr'}
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 302)

    def test_failed_callback(self):
        url = reverse('openid_callback')
        response = self.client.get(url)
        self.assertContains(response, 'Invalid openid.mode')

    def test_successful_callback(self):
        url = reverse('openid_callback') + '?type=success'
        response = self.client.get(url)
        self.assertContains(response, 'OpenID association')



import time

try:
    from openid.consumer.discover import discover
    from openid.extensions import sreg, ax
    from openid.yadis import xri
except ImportError:
    from django.core.exceptions import ImproperlyConfigured
    raise ImproperlyConfigured(
        "python-openid is required to use le_social.openid"
    )

from django.http import get_host
from django.utils.html import escape


class OpenID(object):
    def __init__(self, openid, issued, attrs=None, sreg_=None, ax_=None):
        self.openid = openid
        self.issued = issued
        self.attrs = attrs or {}
        self.sreg = sreg_ or {}
        self.ax = ax_ or {}
        self.is_iname = (xri.identifierScheme(openid) == 'XRI')

    def __repr__(self):
        return '<OpenID: %s>' % self.openid

    def __str__(self):
        return self.openid


def get_url_host(request):
    scheme = 'https' if request.is_secure() else 'http'
    host = escape(get_host(request))
    return '%s://%s' % (scheme, host)


def discover_extensions(openid_url):
    service = discover(openid_url)
    use_ax = False
    use_sreg = False
    for endpoint in service[1]:
        if not use_sreg:
            use_sreg = sreg.supportsSReg(endpoint)
        if not use_ax:
            use_ax = endpoint.usesExtension("http://openid.net/srv/ax/1.0")
        if use_ax and use_sreg:
            break
    if not use_sreg and not use_ax:
        use_sreg = True
    return use_ax, use_sreg


def from_openid_response(openid_response):
    issued = int(time.time())
    sreg_resp = (sreg.SRegResponse.fromSuccessResponse(openid_response) or
                 [])
    ax_resp = ax.FetchResponse.fromSuccessResponse(openid_response)
    ax_args = {}
    if ax_resp is not None:
        ax_args = ax_resp.getExtensionArgs()
        ax_resp.parseExtensionArgs(ax_args)
        ax_args = ax_resp.data

    return OpenID(
        openid_response.identity_url, issued, openid_response.signed_fields,
        dict(sreg_resp), ax_args,
    )



# -*- coding: utf-8 -*-
from distutils.core import setup
from setuptools import find_packages

import le_social


setup(
    name='django-le-social',
    version=le_social.__version__,
    author='Bruno Renié',
    author_email='bruno@renie.fr',
    packages=find_packages(),
    include_package_data=True,
    url='https://github.com/brutasse/django-le-social',
    license='BSD licence, see LICENCE file',
    description='External registration / authentication for Django',
    long_description=open('README.rst').read(),
    install_requires=[
        'Django',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    test_suite='runtests.runtests',
    zip_safe=False,
)



from django.core.urlresolvers import reverse
from django.http import HttpResponse

from .. import views


class Common(object):
    """
    Stuff shared by the two views.
    """
    def get_return_url(self):
        return reverse('openid_callback')

    def failure(self, message):
        return HttpResponse(message)


class Begin(Common, views.Begin):
    template_name = 'le_social/openid/openid.html'
begin = Begin.as_view()


class Callback(Common, views.Callback):
    def success(self):
        openid_url = self.openid_response.identity_url
        return HttpResponse('OpenID association: %s' % openid_url)
callback = Callback.as_view()



from __future__ import absolute_import

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.shortcuts import redirect
from django.views import generic

try:
    from twitter import Twitter, OAuth, TwitterError
    from twitter.oauth_dance import parse_oauth_tokens
except ImportError:
    raise ImproperlyConfigured(
        "twitter<1.8 is required to use le_social.twitter."
    )


class OAuthMixin(object):
    consumer_key = None
    consumer_secret = None

    def get_consumer_key(self):
        if self.consumer_key is not None:
            return self.consumer_key
        if hasattr(settings, 'CONSUMER_KEY'):
            return settings.CONSUMER_KEY
        else:
            raise ImproperlyConfigured("Set settings.CONSUMER_KEY or the "
                                       "consumer_key attribute or "
                                       "implement get_consumer_key")

    def get_consumer_secret(self):
        if self.consumer_secret is not None:
            return self.consumer_secret
        if hasattr(settings, 'CONSUMER_SECRET'):
            return settings.CONSUMER_SECRET
        else:
            raise ImproperlyConfigured("Set settings.CONSUMER_SECRET or the "
                                       "consumer_secret attribute or "
                                       "implement get_consumer_secret")


class Authorize(generic.View, OAuthMixin):
    """
    A base class for the authorize view. Just sets the request token
    in the session and redirects to twitter.
    """
    def get(self, request, force_login=False, *args, **kwargs):
        callback = self.build_callback()
        oauth = OAuth('', '',
                      self.get_consumer_key(),
                      self.get_consumer_secret())
        api = Twitter(auth=oauth, secure=True, format='', api_version=None)
        (oauth.token, oauth.token_secret) = parse_oauth_tokens(
            api.oauth.request_token(oauth_callback=callback))
        request.session['request_token'] = (oauth.token, oauth.token_secret)
        url = ('https://api.twitter.com/oauth/authenticate?oauth_token='
               '%s' % oauth.token)
        if force_login:
            url += '&force_login=true'
        return redirect(url)

    def build_callback(self):
        """ Override this if you'd like to specify a callback URL"""
        return None


class Callback(generic.View, OAuthMixin):
    """
    A base class for the return callback. Subclasses must define:

        - error(error_msg, exception=None): what to do when
          something goes wrong? Must return an HttpResponse

        - success(auth): what to do on successful auth? Do
          some stuff with the twitter.OAuth object and return
          an HttpResponse
    """
    def get(self, request, *args, **kwargs):
        verifier = request.GET.get('oauth_verifier', None)
        if verifier is None:
            return self.error('No verifier code')

        if not 'request_token' in request.session:
            return self.error('No request token found in the session')

        request_token = request.session.pop('request_token')
        request.session.modified = True

        oauth = OAuth(request_token[0], request_token[1],
                      self.get_consumer_key(),
                      self.get_consumer_secret())
        api = Twitter(auth=oauth, secure=True, format='', api_version=None)
        try:
            (oauth.token, oauth.token_secret) = parse_oauth_tokens(
                api.oauth.access_token(oauth_verifier=verifier))
        except TwitterError:
            return self.error('Failed to get an access token')

        return self.success(oauth)

    def success(self, auth):
        """
        Twitter authentication successful, do some stuff with his key.
        """
        raise NotImplementedError("You need to provide an implementation of "
                                  "success(auth)")

    def error(self, message, exception=None):
        """
        Meh. Something broke.
        """
        raise NotImplementedError("You need to provide an implementation of "
                                  "error(message, exception=None)")



from __future__ import absolute_import

import twitter
from twitter import oauth_dance

from django.core.urlresolvers import reverse
from django.test import TestCase


def parse_oauth_tokens(tokens):
    return tokens


class OAuth(object):
    def request_token(self, oauth_callback):
        return 'token', 'token_secret'

    def access_token(self, oauth_verifier):
        return 'Yay', 'access token'


class Account(object):
    def verify_credentials(self):
        return {'screen_name': 'brutasse'}


class Twitter(object):
    def __init__(self, **kwargs):
        for key in kwargs:
            setattr(self, key, kwargs[key])
        self.oauth = OAuth()
        self.account = Account()


class TwitterTest(TestCase):
    urls = 'le_social.twitter.tests.urls'

    def setUp(self):
        self._old_twitter = twitter.Twitter
        twitter.Twitter = Twitter

        self._old_parse_tokens = oauth_dance.parse_oauth_tokens
        oauth_dance.parse_oauth_tokens = parse_oauth_tokens

    def tearDown(self):
        twitter.Twitter = self._old_twitter
        oauth_dance.parse_oauth_tokens = self._old_parse_tokens

    def test_authorize(self):
        url = reverse('authorize')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 302)

    def test_callback(self):
        url = reverse('callback')
        response = self.client.get(url)
        self.assertContains(response, "No verifier code")

        url += '?oauth_verifier=foobar'
        response = self.client.get(url)
        self.assertContains(response, "No request token found in the session")

        self.client.get(reverse('authorize'))
        response = self.client.get(url)
        self.assertContains(response, 'brutasse')



try:
    from openid.consumer.consumer import (Consumer, SUCCESS, CANCEL, FAILURE,
                                          SETUP_NEEDED)
    from openid.consumer.discover import DiscoveryFailure
    from openid.extensions import sreg, ax
except ImportError:
    from django.core.exceptions import ImproperlyConfigured
    raise ImproperlyConfigured(
        "python-openid is required to use le_social.openid"
    )

from django.shortcuts import redirect
from django.utils.encoding import smart_unicode
from django.utils.translation import ugettext_lazy as _
from django.views import generic

from .forms import OpenIDForm
from .middleware import OpenIDMiddleware
from .store import DjangoOpenIDStore
from .utils import get_url_host, discover_extensions, from_openid_response


class ReturnUrlMixin(object):
    return_url = None

    def get_return_url(self):
        if self.return_url is None:
            raise NotImplementedError("Either set the return_url or "
                                      "implement get_return_url()")
        return get_url_host(self.request) + self.return_url


class Begin(generic.FormView, ReturnUrlMixin):
    form_class = OpenIDForm
    sreg_attrs = {}
    ax_attrs = []
    store_class = DjangoOpenIDStore
    trust_root = '/'

    def form_valid(self, form):
        openid_url = form.cleaned_data['openid_url']
        return_url = self.get_return_url()
        return self.ask_openid(openid_url, return_url)

    def get_trust_root(self):
        return get_url_host(self.request) + self.trust_root

    def get_sreg_attrs(self):
        return self.sreg_attrs

    def get_ax_attrs(self):
        return self.ax_attrs

    def ask_openid(self, openid_url, return_url):
        trust_root = self.get_trust_root()
        consumer = Consumer(self.request.session, self.store_class())

        try:
            auth_request = consumer.begin(openid_url)
        except DiscoveryFailure:
            message = _('The OpenID %(url)s was invalid')
            return self.failure(message % {'url': openid_url})

        use_ax, use_sreg = discover_extensions(openid_url)
        sreg_request = None
        ax_request = None
        if use_sreg:
            sreg_attrs = self.get_sreg_attrs()
            if 'optional' not in sreg_attrs:
                sreg_attrs.update({'optional': ['nickname', 'email']})
            sreg_request = sreg.SRegRequest(**sreg_attrs)
        if use_ax:
            ax_request = ax.FetchRequest()
            ax_request.add(
                ax.AttrInfo('http://schema.openid.net/contact/email',
                            alias='email', required=True),
            )
            ax_request.add(
                ax.AttrInfo('http://schema.openid.net/namePerson/friendly',
                            alias='nickname', required=True),
            )
            ax_attrs = self.get_ax_attrs()
            for attr in ax_attrs:
                if len(attr) == 2:
                    ax_request.add(ax.AttrInfo(attr[0], required=attr[1]))
                else:
                    ax_request.add(ax.AttrInfo(attr[0]))

        if sreg_request is not None:
            auth_request.addExtension(sreg_request)
        if ax_request is not None:
            auth_request.addExtension(ax_request)
        redirect_url = auth_request.redirectURL(trust_root, return_url)
        return redirect(redirect_url)

    def failure(self, message):
        raise NotImplementedError("You need to provide an implementation "
                                  "of failure()")


class BadOpenIDStatus(Exception):
    pass


class Callback(generic.View, ReturnUrlMixin):
    def success(self):
        """
        Gets called when the OpenID authentication is successful.
        """
        raise NotImplementedError("You need to provide an implementation "
                                  "of success()")

    def failure(self, message):
        """
        Gets called when the OpenID authentication fails.
        """
        raise NotImplementedError("You need to provide an implementation "
                                  "of failure()")

    def get(self, request, *args, **kwargs):
        consumer = Consumer(request.session, DjangoOpenIDStore())
        query = dict((k, smart_unicode(v)) for k, v in request.GET.items())
        openid_response = consumer.complete(query, self.get_return_url())
        self.openid_response = openid_response

        if openid_response.status == SUCCESS:
            if 'openids' not in request.session.keys():
                request.session['openids'] = []
            request.session['openids'] = filter(
                lambda o: o.openid != openid_response.identity_url,
                request.session['openids'],
            )
            request.session['openids'].append(
                from_openid_response(openid_response),
            )
            OpenIDMiddleware().process_request(request)
            return self.success()
        elif openid_response.status == CANCEL:
            return self.failure(_('The request was cancelled'))
        elif openid_response.status == FAILURE:
            return self.failure(openid_response.message)
        elif openid_response.status == SETUP_NEEDED:
            return self.failure(_('Setup needed'))
        else:
            raise BadOpenIDStatus(openid_response.status)
